from pyspark import SparkContext
from pyspark.sql import SparkSession
from pyspark.sql.functions import col, size, expr, udf, sum, count_distinct
from pyspark.sql.types import StructField, StructType, IntegerType

sc = SparkContext(appName="blunders_per_type")
spark = SparkSession.builder.appName("blunders_per_type").getOrCreate()

################################################# VARIABLES #############################################
beginner_upper_limit = 1000
intermediate_upper_limit = 1400
advanced_upper_limit = 2000

# Schema for UDF return type
metric_schema = StructType([
    StructField("K", IntegerType(), False),
    StructField("Q", IntegerType(), False),
    StructField("R", IntegerType(), False),
    StructField("B", IntegerType(), False),
    StructField("N", IntegerType(), False),
    StructField("P", IntegerType(), False)
    ])

pieces = ["K", "Q", "R", "B", "N", "P"] # Chess pieces, with P for pawn (not-standard)

metrics = spark.read.parquet('hdfs:///user/s2755572/Chess_Metrics/Lichess_202312_CCRL_1990030/') \
    .withColumn("TotalWhiteMoves", size(col('WhiteMetrics.DCM'))) \
    .withColumn("TotalBlackMoves", size(col('BlackMetrics.DCM')))

# Filters the inner array of pgn_move in DCM to only contain moves that are blunders.
# Interesting note, there are no blunders in the computer games ;) -- 27 blunders are "present", but they are just comments made by the analysis engine.
filtered_metrics = metrics.withColumn("WhiteMetricsNEW", expr("filter(WhiteMetrics.DCM, x -> x.pgn_move LIKE '%??%')")) \
    .withColumn("BlackMetricsNEW", expr("filter(BlackMetrics.DCM, x -> x.pgn_move LIKE '%??%')")) \
    .drop("WhiteMetrics", "BlackMetrics", "isHuman") \
    .withColumnRenamed("WhiteMetricsNEW", "WhiteMetrics") \
    .withColumnRenamed("BlackMetricsNEW", "BlackMetrics") \
    .filter((size(col("WhiteMetrics")) > 0) | (size(col("BlackMetrics")) > 0)) \
    .withColumn("WhiteBlunderedMoves", size(col('WhiteMetrics.DCM'))) \
    .withColumn("BlackBlunderedMoves", size(col('BlackMetrics.DCM')))
    

@udf(returnType=metric_schema)
def get_blunder_counts_per_piece(metric_struct):
    """
    Return a struct with the count of blunders per piece type.
    """
    blunder_counts = {"K": 0, "Q": 0, "R": 0, "B": 0, "N": 0, "P": 0}
    for metric in metric_struct:
        piece_type = metric["piece_type"]
        blunder_counts[piece_type] += 1
    return blunder_counts

# Get the blunder counts per piece type for both white and black players.
aggregated_metrics = filtered_metrics.withColumn("WhiteBlunders", get_blunder_counts_per_piece(col("WhiteMetrics"))) \
    .withColumn("BlackBlunders", get_blunder_counts_per_piece(col("BlackMetrics")))

# Classify the games into Beginner, Intermediate, Advanced, and Master categories based on the average rating of the players.
classified_metrics = aggregated_metrics.withColumn("AverageRating", (col("WhiteElo") + col("BlackElo")) / 2) \
    .withColumn("RatingCategory", expr(
    "CASE WHEN AverageRating <= 1000 THEN 'Beginner' WHEN AverageRating <= 1400 THEN 'Intermediate' WHEN AverageRating <= 2000 THEN 'Advanced' ELSE 'Master' END")) \
    .drop(col('WhiteElo'), col('BlackElo'), col('AverageRating'), col('WhiteMetrics'), col('BlackMetrics'))

# Expand the metrics to have a column for each piece type.
expanded_metrics = classified_metrics.select("game_idx", "RatingCategory", "WhiteBlunderedMoves", "BlackBlunderedMoves", "TotalWhiteMoves", "TotalBlackMoves",
                                 *[col(f"WhiteBlunders.{piece}").alias(f"White_{piece}") for piece in pieces],
                                 *[col(f"BlackBlunders.{piece}").alias(f"Black{piece}") for piece in pieces])

# Aggregate the metrics to get the total number of games and blunders per piece type.
results = expanded_metrics.drop('TotalGameMoves') \
    .groupBy("RatingCategory") \
    .agg(
        count_distinct("game_idx").alias("NumberOfGames"),
        *[
            sum(column).alias(column)
            for column in expanded_metrics.columns
            if column not in ["RatingCategory", "game_idx", "TotalGameMoves"]
        ]
    )

# Save the results to a CSV file, repartition 1 to save it to a single output file for easier reading. (it is small data, should be okay)
results.repartition(1).write.csv('file:///home/s2755572/blunders_per_type.csv', header=True)
