import re
from itertools import zip_longest
from pyspark import SparkContext
from pyspark.sql import SparkSession
from pyspark.sql.functions import udf, col, monotonically_increasing_id, lit, arrays_zip, explode, median
from pyspark.sql.types import IntegerType, FloatType, StructType, StructField, StringType, ArrayType

sc = SparkContext(appName="human_computer_metrics")
spark = SparkSession.builder.appName("human_computer_metrics").getOrCreate()

################################################# REGEX #################################################
# Lookaheads do not work since they are not fixed width.
eval_regex_black_only_lookahead_lichess = '(?<=\d+\.\.\.\s\w+[=-]?\w?[-]?\w+[?!#+]*\s\{\s\[%eval\s)(-?\d+\.\d+|-?\d+)(?=\])' # Extract only the numerical evaluations for Black's moves (https://regexr.com/8athn)
eval_regex_white_only_lookahead_lichess = '(?<=\d+\.\s\w+[=-]?\w?[-]?\w+[?!#+]*\s\{\s\[%eval\s)(-?\d+\.\d+|-?\d+)(?=\])' # Extract only the numerical evaluations for White's moves (https://regexr.com/8athq)
eval_regex_black_only_lichess = '(\d+\.{3}\s(\w+[=-]?\w?[-]?\w+[?!#+]*)\s\{\s\[%eval\s(#*-?\d+\.*\d*)\])' # Extract all moves with only the evaluations for Black (https://regexr.com/8ath2)
eval_regex_white_only_lichess = '(\d+\.\s(\w+[=-]?\w?[-]?\w+[?!#+]*)\s\{\s\[%eval\s(#*-?\d+\.*\d*)\])' # Extract all moves with only the evaluations for White (https://regexr.com/8atgs)
eval_regex_numerical_lichess = '(?<=\[%eval\s)(-?\d+\.\d+|-?\d+)' # Extract the numerical value out of a captured eval regex (https://regexr.com/8atht)
eval_regex_numerical_CCRL = '[-+]#?\d+.\d+(?=\/.*})|[-+]\d(?=\/.*})|[-+]#\d+,\d+' # Extract only the numerical value from the moves data in CCRL. Also need to remove # manually... (regexr.com/8attk)
regex_destination_lichess = '([a-h][1-8]|O(?:-O){1,2})' # Extract the destination of a PGN move (https://regexr.com/8aunv)
regex_destination_CCRL = '([^\s]*(?<!\()([a-h][1-8]|O(?:-O){1,2})(?!\w?\)))' # Extract the destination from CCRL's proprietary PGN format (https://regexr.com/8av8l)

################################################# VARIABLES #############################################
checkmate_eval_compensation = 30.0

################################################# UTILS #################################################

def checkmate_eval_to_float(str_eval: str) -> float:
    '''
    Convert a PGN string evaluation to a float.
    If the string contains a '#', it means that there is a forced checkmate in X amount of moves.
    For this, we will add a compensation of {checkmate_eval_compensation} to the evaluation.
    '''
    if ('#' not in str_eval):
        return float(str_eval)
    eval = float(str_eval.replace('#','').replace(',','.'))
    is_positive = False if eval < 0 else True
    return (eval + checkmate_eval_compensation) if is_positive else (eval-checkmate_eval_compensation)

def get_piece_type(pgn_move: str) -> str:
    '''
    Extract the piece type from a PGN notation move.

    @Example:
    Nf3 will return 'N'
    e4 will return 'P'
    O-O will return 'K,R' (it involves two pieces)
    '''
    if pgn_move in ['O-O', 'O-O-O']:
        return 'K,R'
    if '=' in pgn_move:
        return 'P'
    match = re.match('[KQRBN]', pgn_move)
    if match is not None:
        return match.group(0)
    return 'P'
    

################################################# UDF ###################################################

# Metric schema used to define a return type for the UDFs
metric_schema = StructType([
    StructField("ACPL", FloatType(), False),
    StructField("DCM", ArrayType(StructType([
        StructField("move", IntegerType(), False),
        StructField("pgn_move", StringType(), True),
        StructField("piece_type", StringType(), True),
        StructField("destination", StringType(), True),
        StructField("dcm", FloatType(), False)
    ])), False)
])

@udf(returnType=IntegerType())
def total_time(tc: str):
    """
    According to Lichess, the total time of a game is estimated as follows:
    Lichess time controls are based on estimated game duration = (clock initial time in seconds) + 40 × (clock increment).
    For instance, the estimated duration of a 5+3 game is 5 × 60 + 40 × 3 = 420 seconds.
    """
    if tc is not None:
        initial_time = tc.split('+')[0]
        clk_increment = tc.split('+')[1]
        return int(initial_time) + 40 * int(clk_increment)
    
@udf(returnType=metric_schema)
def compute_metrics_white_lichess(eval_str: str):
    """
    Compute the following metrics for the White player:
    {
        "ACPL": float - Average centipawn loss per whole match.
        "DCM": array[{
                "move": integer - Move number from PGN
                "destination": string - Destination square of the move in PGN
                "dcm" float - Delta of change between White's move and Black's last reply (did he capitalize?).
            }
    }
    """
    if eval_str is not None:
        white_moves     = re.findall(eval_regex_white_only_lichess, eval_str)
        black_moves     = re.findall(eval_regex_black_only_lichess, eval_str)
        white_evals     = [checkmate_eval_to_float(value[2]) for value in white_moves]
        black_evals     = [checkmate_eval_to_float(value[2]) for value in black_moves]
        pgn_moves       = [value[1] for value in white_moves]
        piece_types     = [get_piece_type(move) for move in pgn_moves]
        eval_pairs      = [move_pair for move_pair in tuple(zip_longest(white_evals, black_evals))]
        destinations    = [value for index, value in enumerate(re.findall(regex_destination_lichess, eval_str)) if index % 2 == 0]
        acpl            = 0.0
        dcm             = []
        for i in range(len(eval_pairs)):
            if (i == 0):
                dcm.append({
                    "move": i+1,
                    "pgn_move": pgn_moves[i] if i < len(pgn_moves) else None,
                    "piece_type": piece_types[i] if i < len(piece_types) else None,
                    "destination": destinations[i] if i < len(destinations) else None,
                    "dcm": 0.0
                })
                continue
            eval_first_move = eval_pairs[i-1][1]
            eval_second_move = eval_pairs[i][0]
            dcm.append({
                "move": i+1,
                "pgn_move": pgn_moves[i] if i < len(pgn_moves) else None,
                "piece_type": piece_types[i] if i < len(piece_types) else None,
                "destination": destinations[i] if i < len(destinations) else None,
                "dcm": eval_second_move - eval_first_move
            })
            acpl += (eval_second_move - eval_first_move)
        acpl = acpl / len(white_evals) if len(white_evals) > 0 else 0.0
        return {
            "ACPL": acpl,
            "DCM": dcm
        }
    
@udf(returnType=metric_schema)
def compute_metrics_black_lichess(eval_str: str):
    """
    Compute the following metrics for the Black player:
    {
        "ACPL": float - Average centipawn loss per whole match.
        "DCM": array[{
                "move": integer - Move number from PGN
                "destination": string - Destination square of the move in PGN
                "dcm" float - Delta of change between Black's move and White's last reply (did he capitalize?) If Black's DCM is positive, he did a mistake, because white's advantage increased.
            }
    }
    implNote: Lichess has a more standard FIDE algebraic notation, although sometimes different from the CCRL dataset, meaning that
    two different REGEX styles have to be used for the two datasets.
    """
    if eval_str is not None:
        white_moves     = re.findall(eval_regex_white_only_lichess, eval_str)
        black_moves     = re.findall(eval_regex_black_only_lichess, eval_str)
        white_evals     = [checkmate_eval_to_float(value[2]) for value in white_moves]
        black_evals     = [checkmate_eval_to_float(value[2]) for value in black_moves]
        pgn_moves       = [value[1] for value in black_moves]
        piece_types     = [get_piece_type(move) for move in pgn_moves]
        eval_pairs      = [move_pair for move_pair in tuple(zip_longest(white_evals, black_evals))]
        destinations    = [value for index, value in enumerate(re.findall(regex_destination_lichess, eval_str)) if index % 2 == 1]
        acpl            = 0.0
        dcm             = []
        for i in range(len(eval_pairs)):
            eval_first_move = eval_pairs[i][0]
            eval_second_move = eval_pairs[i][1]
            if (eval_second_move == None):
                break
            dcm.append({
                "move": i+1,
                "pgn_move": pgn_moves[i] if i < len(pgn_moves) else None,
                "piece_type": piece_types[i] if i < len(piece_types) else None,
                "destination": destinations[i] if i < len(destinations) else None,
                "dcm": eval_second_move - eval_first_move
            })
            acpl += (eval_second_move - eval_first_move)
        acpl = acpl / len(black_evals) if len(black_evals) > 0 else 0.0
        return {
            "ACPL": acpl,
            "DCM": dcm
        }
    
@udf(returnType=metric_schema)
def compute_metrics_white_ccrl(eval_str: str):
    """
    Compute the following metrics for the White player:
    {
        "ACPL": float - Average centipawn loss per whole match.
        "DCM": array[{
                "move": integer - Move number from PGN
                "destination": string - Destination square of the move in PGN
                "dcm" float - Delta of change between White's move and Black's last reply (did he capitalize?)
            }
    }
    """
    if eval_str is not None:
        evaluations     = re.findall(eval_regex_numerical_CCRL, eval_str)
        white_evals     = [checkmate_eval_to_float(value) for index, value in enumerate(evaluations) if index % 2 == 0]
        black_evals     = [checkmate_eval_to_float(value) for index, value in enumerate(evaluations) if index % 2 == 1]
        eval_pairs      = [move_pair for move_pair in tuple(zip_longest(white_evals, black_evals))]
        pgn_moves       = [value[0] for index, value in enumerate(re.findall(regex_destination_CCRL, eval_str)) if index % 2 == 0]
        piece_types     = [get_piece_type(value) for value in pgn_moves]
        destinations    = [value[1] for index, value in enumerate(re.findall(regex_destination_CCRL, eval_str)) if index % 2 == 0]
        acpl            = 0.0
        dcm             = []
        for i in range(len(eval_pairs)):
            if (i == 0):
                dcm.append({
                    "move": i+1,
                    "pgn_move": pgn_moves[i] if i < len(pgn_moves) else None,
                    "piece_type": piece_types[i] if i < len(piece_types) else None,
                    "destination": destinations[i] if i < len(destinations) else None,
                    "dcm": 0.0
                })
                continue
            eval_first_move = eval_pairs[i-1][1]
            eval_second_move = eval_pairs[i][0]
            dcm.append({
                "move": i+1,
                "pgn_move": pgn_moves[i] if i < len(pgn_moves) else None,
                "piece_type": piece_types[i] if i < len(piece_types) else None,
                "destination": destinations[i] if i < len(destinations) else None,
                "dcm": eval_second_move - eval_first_move
            })
            acpl += (eval_second_move - eval_first_move)
        acpl = acpl / len(white_evals) if len(white_evals) > 0 else 0.0
        return {
            "ACPL": acpl,
            "DCM": dcm
        }

@udf(returnType=metric_schema)
def compute_metrics_black_ccrl(eval_str: str):
    """
    Compute the following metrics for the Black player:
    {
        "ACPL": float - Average centipawn loss per whole match.
        "DCM": array[{
                "move": integer - Move number from PGN
                "destination": string - Destination square of the move in PGN
                "dcm" float - Delta of change between Black's move and White's last reply (did he capitalize?) If Black's DCM is positive, he did a mistake, because white's advantage increased.
            }
    }
    implNote: Lichess has a more standard FIDE algebraic notation, although sometimes different from the CCRL dataset, meaning that
    two different REGEX styles have to be used for the two datasets.
    known_issues: Since we compute pairs of moves, whenever a last move is 
    """
    if eval_str is not None:
        evaluations     = re.findall(eval_regex_numerical_CCRL, eval_str)
        white_evals     = [checkmate_eval_to_float(value) for index, value in enumerate(evaluations) if index % 2 == 0]
        black_evals     = [checkmate_eval_to_float(value) for index, value in enumerate(evaluations) if index % 2 == 1]
        eval_pairs      = [move_pair for move_pair in tuple(zip_longest(white_evals, black_evals))]
        pgn_moves       = [value[0] for index, value in enumerate(re.findall(regex_destination_CCRL, eval_str)) if index % 2 == 1]
        piece_types     = [get_piece_type(value) for value in pgn_moves]
        destinations    = [value[1] for index, value in enumerate(re.findall(regex_destination_CCRL, eval_str)) if index % 2 == 1]
        acpl            = 0.0
        dcm             = []
        for i in range(len(eval_pairs)):
            eval_first_move = eval_pairs[i][0]
            eval_second_move = eval_pairs[i][1]
            if (eval_second_move == None):
                break
            dcm.append({
                "move": i+1,
                "pgn_move": pgn_moves[i] if i < len(pgn_moves) else None,
                "piece_type": piece_types[i] if i < len(piece_types) else None,
                "destination": destinations[i] if i < len(destinations) else None,
                "dcm": eval_second_move - eval_first_move
            })
            acpl += (eval_second_move - eval_first_move)
        acpl = acpl / len(black_evals) if len(black_evals) > 0 else 0.0
        return {
            "ACPL": acpl,
            "DCM": dcm
        }

"""
Metric calculation for human and computer games.
The metrics are computed for both White and Black players.
"""
human = spark.read.parquet('hdfs:///data/doina/Lichess/lichess_db_standard_rated_2023-12.parquet') \
    .withColumn('game_idx', monotonically_increasing_id()) \
    .withColumn('isHuman', lit(True)) \
    .withColumn('TotalTime', total_time('TimeControl')) \
    .filter(col('Moves').contains('%eval')) \
    .filter(col("Event").contains('Rapid') | col("Event").contains('Classical')) \
    .withColumn('WhiteMetrics', compute_metrics_white_lichess('Moves')) \
    .withColumn('BlackMetrics', compute_metrics_black_lichess('Moves')) 

computer = spark.read.option('header', True).csv('hdfs:///user/s2755572/Computer_Chess_1990030_Parsed/*') \
    .withColumn('isHuman', lit(False)) \
    .withColumn('WhiteMetrics', compute_metrics_white_ccrl('data')) \
    .withColumn('BlackMetrics', compute_metrics_black_ccrl('data')) 

# Moves can be removed after validating their correctness
computer_metrics = computer.select(col('game_idx'), col('WhiteElo'), col('BlackElo'), col('isHuman'), col('WhiteMetrics'), col('BlackMetrics'))
human_metrics = human.select(col('game_idx'), col('WhiteElo'), col('BlackElo'), col('isHuman'), col('WhiteMetrics'), col('BlackMetrics'))

""" Metrics will be exported as JSON files since we need more flexible schemas (nested) which cannot be stored in CSV format"""
metrics =  human_metrics.union(computer_metrics)
metrics.repartition(120).write.parquet(path='hdfs:///user/s2755572/Chess_Metrics/Lichess_202312_CCRL_1990030/', mode='overwrite')
